# Copyright (c) 2017-2019 Uber Technologies, Inc.
# SPDX-License-Identifier: Apache-2.0

import warnings

from pyroapi import *  # noqa F401
from pyroapi import __all__  # noqa F401

warnings.warn("pyro.generic has moved to the pyroapi package", DeprecationWarning)
